#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>

#include <unistd.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include "speckeysd.h"

Display * dpy;			/* The connection to the X server. */
int screen_count;		/* The number of screens. */

char * argv0;

HotKey * hotkeys = NULL;

void panic(char *);
void keypress(XEvent *);
void read_hot_key_file(char *);

int
main(int argc, char *argv[])
{
	XEvent ev;
	
	argv0 = argv[0];
	
	/* Open a connection to the X server. */
	dpy = XOpenDisplay("");
	if (dpy == 0)
		panic("can't open display.");
	
	/* Set up signal handlers. */
#if 0
	signal(SIGTERM, SIG_IGN);
	signal(SIGINT, SIG_IGN);
#endif
	signal(SIGHUP, SIG_IGN);
	
	screen_count = ScreenCount(dpy);
	
	if (argc != 2)
		panic("syntax: speckeysd <keys file>");
	
	read_hot_key_file(argv[1]);
	
	/* Make sure all our communication to the server got through. */
	XSync(dpy, False);
	
	/* The main event loop. */
	for (;;) {
		XNextEvent(dpy, &ev);
		switch (ev.type) {
		case KeyPress:
			keypress(&ev);
			break;
		case MappingNotify:
			XRefreshKeyboardMapping((XMappingEvent *) &ev);
			break;
		default:
			/* Do I look like I care? */
			break;
		}
	}
}

unsigned int
parse_modifiers(char * name)
{
	char* separator = strchr(name, '-');
	unsigned int modifiers = 0;
	if (separator != NULL)
	{
	    *separator = 0;
	    modifiers |= parse_modifiers(separator+1);
	}
	if (!strcmp(name, "Shift"))
	    modifiers |= ShiftMask;
	else if (!strcmp(name, "Control"))
	    modifiers |= ControlMask;
	else if (!strcmp(name, "Alt"))
	    modifiers |= Mod1Mask;
	return modifiers;
}

void
add_hot_key(char * keyname, char * command)
{
	int screen;
	HotKey * new_key;
	char* unmodified = strrchr(keyname, '-');
	unsigned int modifiers = 0;

	if (unmodified == NULL) {
	    unmodified = keyname;
    } else {
        *unmodified = 0;
        ++ unmodified;
        modifiers = parse_modifiers(keyname);
    }
	
	new_key = (HotKey *) malloc(sizeof(HotKey));
	new_key->keysym = XStringToKeysym(unmodified);
	new_key->modifiers = modifiers;
	new_key->command = strdup(command);
	new_key->next = hotkeys;
	hotkeys = new_key;
	
	for (screen = 0; screen < screen_count; screen++) {
		Window root = RootWindow(dpy, screen);
		XGrabKey(dpy, XKeysymToKeycode(dpy, new_key->keysym),
			modifiers, root, False, GrabModeAsync, GrabModeAsync);
	}
}

void
read_hot_key_file(char * fname)
{
	FILE * fp;
	char buf[BUFSIZ];
	
	fp = fopen(fname, "r");
	if (fp == NULL) {
		fprintf(stderr, "%s: couldn't read \"%s\".\n", argv0, fname);
		return;
	}
	
	while (fgets(buf, BUFSIZ, fp) != NULL) {
		char * tab = strchr(buf, '\t');
		if (*buf == '#' || tab == NULL)
			continue;
		
		*tab = 0;
		add_hot_key(buf, tab+1);
	}
	
	fclose(fp);
}

void
keypress(XEvent * ev0)
{
    XKeyEvent * ev = (XKeyEvent *) ev0;
	char buffer[1];
	KeySym keysym;
	int c;
	HotKey * h = hotkeys;
	
	c = XLookupString(ev, buffer, sizeof(buffer), &keysym, 0);
	for (; h != NULL; h = h->next) {
		if (h->keysym == keysym && ev->state == h->modifiers) {
			shell(h->command);
			break;
		}
	}
}

void
reap_zombies()
{
	int status;
	while (waitpid(-1, &status, WNOHANG) > 0);
}

void
shell(char * command)
{
	char * sh;
	
	sh = getenv("SHELL");
	if (sh == 0)
		sh = "/bin/sh";
	
	switch (fork()) {
	case 0:		/* Child. */
		close(ConnectionNumber(dpy));
		execl(sh, sh, "-c", command, 0);
		fprintf(stderr, "%s: can't exec \"%s -c %s\"\n", argv0, sh,
			command);
		exit(EXIT_FAILURE);
	case -1:	/* Error. */
		fprintf(stderr, "%s: couldn't fork\n", argv0);
		break;
	}
	reap_zombies();
}

void
panic(char *s)
{
	fprintf(stderr, "%s: %s\n", argv0, s);
	exit(EXIT_FAILURE);
}
